/* This file contains the procedures for the handling of select
 *
 * The entry points into this file are
 *   do_select:	perform the SELECT system call
 */

#include "fs.h"
#include <minix/com.h>
#include <sys/asynchio.h>	/* for tv */
#include <string.h>		/* for bzero() */
#include <time.h>
#include "file.h"
#include "fproc.h"
#include "inode.h"

FORWARD _PROTOTYPE( void sel_revive, (int proc, int fdn, int oper)	);
FORWARD _PROTOTYPE( int rw_user, (int s, int u, vir_bytes vir,
			 vir_bytes bytes, char *buff, int direction)	);

/* bitfields for minors 0..255 , in and out;
 * used to tell device task which channel to test
 */
PRIVATE fd_mask tty_bits[2][howmany(256, NFDBITS)];
PRIVATE fd_mask net_bits[2][howmany(256, NFDBITS)];
PRIVATE int sel_id = 0;

#define INIT_PROC_NR 2
#define INET_PROC_NR 5

/*===========================================================================*
 *				sel_ack  				     *
 *===========================================================================*/
PUBLIC void sel_ack(who, mptr)
int who;
message *mptr;
{
register struct inode *inoptr;
register struct fproc *lfp;
dev_t dev;

  /* message from task or clock */
  if (mptr->PROC_NR <= 0 || mptr->PROC_NR >= NR_PROCS ||
      (lfp = &fproc[mptr->PROC_NR])->fp_sel_id == 0) return;
  switch (who) {
    case TTY :
      dev = (4 << MAJOR) | (mptr->DEVICE << MINOR);
      break;
    case INET_PROC_NR :
      dev = (7 << MAJOR) | (mptr->DEVICE << MINOR);
      break;
    case CLOCK :
      if (lfp->fp_sel_id != mptr->COUNT) return;
      sel_revive(mptr->PROC_NR, 0, 0);
      return;
    default :  
      return;
  }
  if (lfp->fp_sel_tty >= 0 && lfp->fp_tty == dev) {
      sel_revive(mptr->PROC_NR, lfp->fp_sel_tty, mptr->COUNT);
      return;
  }	
  for (inoptr = inode; inoptr < &inode[NR_INODES]; inoptr++)
    if (inoptr->i_sel_proc == mptr->PROC_NR &&
        (inoptr->i_mode & I_TYPE) == I_CHAR_SPECIAL &&
        (dev_t) inoptr->i_zone[0] == dev) {
		wake_select(inoptr, mptr->COUNT);
		break;
    }
}


/*===========================================================================*
 *				do_select				     *
 *===========================================================================*/
PUBLIC int do_select()
{
/* Perform the select(nd, in, out, ex, tv) system call. */
fd_set in, out, ex;
struct timeval tv, *tvp;
unsigned long time;
fd_mask mask, *ip = (fd_mask *)&in, *op = (fd_mask *)&out, *ep = (fd_mask *)&ex;
struct filp *filptr;
struct inode *inoptr;
u16_t i_mode;
dev_t dev;
int oper, cum = 0, min, tty_cnt = 0, net_cnt = 0;
register i;
#ifdef LOG_SELECT
	printf("fs: do_select() called\n\r");
	printf("fs: do_select(): who = %d\n\r", who);
#endif
  if (who < INIT_PROC_NR || who == INET_PROC_NR) { /* message from task or clock */
	dont_reply = TRUE;
	sel_ack(who, &m);
#ifdef LOG_SELECT
	printf("fs: do_select(): from task or clock\n\r");
#endif
	return OK;
  } 
  /* else : user call SELECT */
  tvp = (struct timeval *) m.m1_i2;
  if (tvp) {
    if (rw_user(D, who, (vir_bytes) tvp, (vir_bytes) sizeof(struct timeval),
                (char *) &tv, WRITING) != OK) return EFAULT;
    time = tv.tv_sec * HZ + (tv.tv_usec * HZ) / 1000000;
    if (time >= LONG_MAX) time = LONG_MAX - 1L;
  } else time = LONG_MAX; /* indefinite timeout */
  if (m.m1_p1) {
    if (rw_user(D, who, (vir_bytes) m.m1_p1, (vir_bytes) sizeof(fd_set),
                (char *) &in, WRITING) != OK) return EFAULT;
  } else FD_ZERO(&in);
  if (m.m1_p2) {
    if (rw_user(D, who, (vir_bytes) m.m1_p2, (vir_bytes) sizeof(fd_set),
                (char *) &out, WRITING) != OK) return EFAULT;
  } else FD_ZERO(&out);
  if (m.m1_p3) {
    if (rw_user(D, who, (vir_bytes) m.m1_p3, (vir_bytes) sizeof(fd_set),
                (char *) &ex, WRITING) != OK) return EFAULT;
  } else FD_ZERO(&ex);

  fp->fp_sel_tty = -1;	/* assume: no /dev/tty */
  mask = 1;
  bzero((char *)tty_bits, howmany(256 * 2, NBBY));
  bzero((char *)net_bits, howmany(256 * 2, NBBY));
  for (i = 0; i < m.m1_i1; i++) {
    oper  = (*ip & mask)? SEL_IN : 0;
    if (*op & mask) oper += SEL_OUT;
    if (*ep & mask) oper += SEL_EX;
    
    if (oper) { /* test this fd */
      if ((filptr = get_filp(i)) == NIL_FILP) {
        cum = err_code;
        break;
      }  
      inoptr = filptr->filp_ino;
      inoptr->i_sel_proc = who;
      inoptr->i_sel_fd = i;
      i_mode = inoptr->i_mode & I_TYPE;
      if (inoptr->i_pipe) {
        pipe_select(inoptr, &oper, filptr->filp_pos);
        if ((oper & SEL_IN ) == 0) *ip &= ~mask; /* reset in flag */
        if ((oper & SEL_OUT) == 0) *op &= ~mask; /* reset out flag */
        if ((oper & SEL_EX ) == 0) *ep &= ~mask; /* reset ex flag */
	/* GH, count active fds */
	if (*ip & mask) cum++;
	if (*op & mask) cum++;
	if (*ep & mask) cum++;
      } else if (i_mode == I_CHAR_SPECIAL) {
        /* cumulate requests of char special inodes in extra fields
         * and examine them later
         */
        dev = (dev_t) inoptr->i_zone[0];
        switch ((dev >> MAJOR) & BYTE) {
	  case 7:	/* /dev/tcp, /dev/udp, /dev/ip, /dev/eth */
            min = (dev >> MINOR) & BYTE;
            if (oper & SEL_IN ) FD_SET(min, (fd_set *)net_bits[0]);
            if (oper & SEL_OUT) FD_SET(min, (fd_set *)net_bits[1]);
            net_cnt++;
	    break;
          case 5:	/* /dev/tty */
            dev = fp->fp_tty; /* fall into */
            fp->fp_sel_tty = i;
          case 4:	/* /dev/ttyXYZ */
            if (dev) {
              min = (dev >> MINOR) & BYTE;
              if (oper & SEL_IN ) FD_SET(min, (fd_set *)tty_bits[0]);
              if (oper & SEL_OUT) FD_SET(min, (fd_set *)tty_bits[1]);
              tty_cnt++;
            } else { /* /dev/tty = /dev/null */
              *ep &= ~mask;
            }
            break;
          default :
            *ep &= ~mask;
        }
      } else { /* normal file or block special */
        /* default : read, write OK, no exception */
        *ep &= ~mask;
      }
    }
    
    mask <<= 1;
    if (mask == 0) {
      mask = 1;
      ip++; op++; ep++;
    }
  } /* for */
  if (net_cnt) { /* send request to INET task */
    if (task_select(7, (fd_set *)net_bits, time ? who : 0) < 0) cum = EBADF;
  }  
  if (tty_cnt) { /* send request to TTY task */
    if (task_select(4, (fd_set *)tty_bits, time ? who : 0) < 0) cum = EBADF;
  }  
    
  if (cum == 0 && (tty_cnt || net_cnt)) {  
    /* examine bitfields of outstanding char special devices */
    mask = 1;
    ip = (fd_mask *)&in; op = (fd_mask *)&out; ep = (fd_mask *)&ex;
    for (i = 0; i < m.m1_i1; i++) {
      oper  = (*ip & mask)? SEL_IN : 0;
      if (*op & mask) oper += SEL_OUT;
      if (*ep & mask) oper += SEL_EX;
      
      if (oper) { /* test this fd */
        if ((filptr = get_filp(i)) == NIL_FILP) {
          cum = err_code;
          break;
        }
        inoptr = filptr->filp_ino;
        i_mode = inoptr->i_mode & I_TYPE;
        if (i_mode == I_CHAR_SPECIAL) {
          dev = (dev_t) inoptr->i_zone[0];
          switch ((dev >> MAJOR) & BYTE) {
	    case 7:	/* /dev/ip */
              min = (dev >> MINOR) & BYTE;
              if (oper & SEL_IN ) {
                if (FD_ISSET(min, (fd_set *)net_bits[0]) == 0) *ip &= ~mask;
              }
              if (oper & SEL_OUT) {
                if (FD_ISSET(min, (fd_set *)net_bits[1]) == 0) *op &= ~mask;
              }
              *ep &= ~mask;  /* no exception */
              break;
            case 5: /* /dev/tty */
              dev = fp->fp_tty; /* fall into */
            case 4: /* /dev/ttyXYZ */
              if (dev) {
                min = (dev >> MINOR) & BYTE;
                if (oper & SEL_IN ) {
                  if (FD_ISSET(min, (fd_set *)tty_bits[0]) == 0) *ip &= ~mask;
                }
                if (oper & SEL_OUT) {
                  if (FD_ISSET(min, (fd_set *)tty_bits[1]) == 0) *op &= ~mask;
                }
                *ep &= ~mask;  /* no exception */
                break;
              } 
          }
        } /* others than char special are allready done */
        /* count active descriptors */
        if (*ip & mask) cum++;
        if (*op & mask) cum++;
        if (*ep & mask) cum++;
      }
    
      mask <<= 1;
      if (mask == 0) {
        mask = 1;
        ip++; op++; ep++;
      }
    } /* for */
  } /* else : no char special in list */
   
  if (cum || time == 0L) { 
    /* remove select marks */
    for (inoptr = inode; inoptr < &inode[NR_INODES]; inoptr++)
      if (inoptr->i_sel_proc == who) inoptr->i_sel_proc = 0;
    if (m.m1_p1) {
      if (rw_user(D, who, (vir_bytes) m.m1_p1, (vir_bytes) sizeof(fd_set),
                  (char *) &in, READING) != OK) cum = EFAULT;
    }
    if (m.m1_p2) {
      if (rw_user(D, who, (vir_bytes) m.m1_p2, (vir_bytes) sizeof(fd_set),
                  (char *) &out, READING) != OK) cum = EFAULT;
    }
    if (m.m1_p3) {
      if (rw_user(D, who, (vir_bytes) m.m1_p3, (vir_bytes) sizeof(fd_set),
                  (char *) &ex, READING) != OK) cum = EFAULT;
    }
#ifdef LOG_SELECT
		printf("fs: do_select(): return(cum)\n\r");
#endif
    return(cum);
  } else {
    fp->fp_sel_in = (vir_bytes) m.m1_p1;
    fp->fp_sel_out = (vir_bytes) m.m1_p2;
    fp->fp_sel_ex = (vir_bytes) m.m1_p3;
    if (++sel_id == 0) sel_id++;
    fp->fp_sel_id = sel_id;
#ifdef LOG_SELECT
		printf("fs: do_select(): suspend(XSELECT)\n\r");
#endif
    suspend(XSELECT);
    if (time < LONG_MAX) clock_select(time);
  }
}

/*===========================================================================*
 *				sel_revive				     *
 *===========================================================================*/
PRIVATE void sel_revive(proc_nr, fdn, oper)
int proc_nr;
int fdn;
int oper;
{
struct fproc *lfp;
fd_set fset;
vir_bytes vir_set = (vir_bytes) 0;
struct inode *inoptr;

  /* remove select marks */
  for (inoptr = inode; inoptr < &inode[NR_INODES]; inoptr++)
    if (inoptr->i_sel_proc == proc_nr) inoptr->i_sel_proc = 0;
  if (proc_nr < INIT_PROC_NR || proc_nr >= NR_PROCS) return;
  FD_ZERO(&fset);
  lfp = &fproc[proc_nr];
  if (lfp->fp_sel_id == 0) return;
  lfp->fp_sel_id = 0;
  if (lfp->fp_sel_in && oper != SEL_IN) {
    if (rw_user(D, proc_nr, lfp->fp_sel_in, (vir_bytes) sizeof(fd_set),
                (char *) &fset, READING) != OK) return;
  }
  if (lfp->fp_sel_out && oper != SEL_OUT) {
    if (rw_user(D, proc_nr, lfp->fp_sel_out, (vir_bytes) sizeof(fd_set),
                (char *) &fset, READING) != OK) return;
  }
  if (lfp->fp_sel_ex && oper != SEL_EX) {
    if (rw_user(D, proc_nr, lfp->fp_sel_ex, (vir_bytes) sizeof(fd_set),
                (char *) &fset, READING) != OK) return;
  }
  if (oper) {
    FD_SET(fdn, &fset);
    switch (oper) {
      case SEL_IN :
        if (lfp->fp_sel_in) vir_set = lfp->fp_sel_in;
        break;
      case SEL_OUT :
        if (lfp->fp_sel_out) vir_set = lfp->fp_sel_out;
        break;
      case SEL_EX :
        if (lfp->fp_sel_ex) vir_set = lfp->fp_sel_ex;
    }
    if (vir_set) {
      if (rw_user(D, proc_nr, vir_set, (vir_bytes) sizeof(fd_set),
                  (char *) &fset, READING) != OK) return;
    }              
  }
  revive( proc_nr, oper ? 1 : 0); /* number of ready fdescs */
}


/*===========================================================================*
 *				wake_select				     *
 *===========================================================================*/
void wake_select(ip, oper)  /* asserts that i_sel_proc != 0 */
struct inode *ip;
int oper;
{
struct fproc *lfp;
int i;

  if (ip->i_sel_proc < INIT_PROC_NR || ip->i_sel_proc >= NR_PROCS) return;
  lfp = &fproc[ip->i_sel_proc];
  switch (oper) {
    case SEL_IN :
      if (!lfp->fp_sel_in) return;
      break;
    case SEL_OUT :
      if (!lfp->fp_sel_out) return;
      break;
    case SEL_EX :
      if (!lfp->fp_sel_ex) return;
  }
  sel_revive(ip->i_sel_proc, ip->i_sel_fd, oper);
}

/*===========================================================================*
 *				rw_user					     *
 *===========================================================================*/
PRIVATE int rw_user(s, u, vir, bytes, buff, direction)
int s;				/* D or T space (stack is also D) */
int u;				/* process number to r/w (usually = 'who') */
vir_bytes vir;			/* virtual address to move to/from */
vir_bytes bytes;		/* how many bytes to move */
char *buff;			/* pointer to FS space */
int direction;			/* TO_USER or FROM_USER */
{
  register int r;

/* Transfer a block of data.  Two options exist, depending on 'direction':
 *     READING:     Move from FS space to user virtual space
 *     WRITING:     Move from user virtual space to FS space
 */

  if (direction == READING ) {
	/* Write from FS space to user space. */
	r = sys_copy(FS_PROC_NR, D, (phys_bytes)buff, u, s, 
		(phys_bytes)vir, (phys_bytes)bytes);
  } else {
	/* Read from user space to FS space. */
	r = sys_copy(u, s, (phys_bytes)vir, FS_PROC_NR, D,
		(phys_bytes)buff, (phys_bytes)bytes);
  }
  return(r);
}
